# encoding: utf-8
# Copyright (C) 2015 John Törnblom
'''
Evaluation of syntax trees constructed from the rule-specification language (RSL).  
'''


import sys
import os
import logging
import traceback

import xtuml.tools

from . import ast
from . import symtab
from . import parse

logger = logging.getLogger(__name__)


class BreakException(Exception):
    pass


class EvalWalker(xtuml.tools.Walker):
    
    def __init__(self, rt, includes):
        xtuml.tools.Walker.__init__(self)
        self.runtime = rt
        self.includes = includes
        self.callstack = list()
        self.symtab = symtab.SymbolTable()
        
        self.symtab.install_global('true', True)
        self.symtab.install_global('false', False)
        self.symtab.install_global('info', self.runtime.info)
                
        self.symtab.enter_scope()
                    
    def accept(self, node, **kwargs):
        self.runtime.info.arch_file_name = node.filename
        self.runtime.info.arch_file_line = node.lineno

        try:
            return xtuml.tools.Walker.accept(self, node, **kwargs)
        except BreakException as e:
            raise e
        except Exception as e:
            print('Traceback  (most recent call last):')
            for n in self.callstack + [node]:
                print('    File "%s", line %d' % (n.filename, n.lineno))
            print(traceback.format_exc())
            sys.exit(e)
    
    def default_accept(self, node, **kwargs):
        print ('> %s' % node.__class__.__name__)
            
    def accept_BodyNode(self, node):
        assert isinstance(node, ast.BodyNode)
        
        self.accept(node.statement_list)
        
    def accept_FunctionNode(self, node):
        assert isinstance(node, ast.FunctionNode)
        
        def _(fn, *args):
            self.symtab.enter_scope()
            
            self.accept(fn.parameter_list, args=reversed(args))
            for stmt in fn.statement_list.statements:
                self.accept(stmt)

            return self.symtab.leave_scope()
        
        self.runtime.define_function(node.name, lambda *args: _(node, *args))
    
    def accept_ParameterListNode(self, node, args):
        assert isinstance(node, ast.ParameterListNode)
        
        for param, arg in zip (node.parameters, args):
            self.accept(param, arg=arg)

    def accept_ParameterNode(self, node, arg):
        assert isinstance(node, ast.ParameterNode)
        
        self.runtime.assert_type(node.type, arg)
        self.symtab.install_symbol(node.name, arg)

    def accept_InvokeNode(self, node):
        assert isinstance(node, ast.InvokeNode)

        args = self.accept(node.argument_list)
        args = [arg.fget() for arg in args]
        
        self.callstack.append(node)
        value = self.runtime.invoke_function(node.function_name, args)
        self.callstack.pop()
        
        if node.variable_name:
            self.symtab.install_symbol(node.variable_name, value)
    
    def accept_ArgumentListNode(self, node):
        assert isinstance(node, ast.ArgumentListNode)
        
        return [self.accept(arg) for arg in node.arguments]

    def accept_StatementListNode(self, node):
        assert isinstance(node, ast.StatementListNode)
        
        for stmt in node.statements:
            self.accept(stmt)
            
    def accept_AssignNode(self, node):
        assert isinstance(node, ast.AssignNode)
        
        value = self.accept(node.expr).fget()
        variable = self.accept(node.variable, mode='w')
        variable.fset(value)
        
    def accept_StringBodyNode(self, node):
        assert isinstance(node, ast.StringBodyNode)
        
        s = ''
        for value in node.values:
            s += self.accept(value).fget()
            
        return property(lambda: s)
    
    def accept_StringValueNode(self, node):
        assert isinstance(node, ast.StringValueNode)
        
        s = node.value
        s = s.replace('\\n', '\n')
        s = s.replace('\\t', '\t')
        
        return property(lambda: s)
        
    def accept_IntegerValueNode(self, node):
        assert isinstance(node, ast.IntegerValueNode)
        
        i = int(node.value)
        
        return property(lambda: i)
        
    def accept_RealValueNode(self, node):
        assert isinstance(node, ast.RealValueNode)
        
        i = float(node.value)
        
        return property(lambda: i)
        
    def accept_VariableAccessNode(self, node, mode='r'):
        assert isinstance(node, ast.VariableAccessNode)
        
        fget = fset = None
        if 'r' in mode:
            value = self.symtab.find_symbol(node.name)
            fget = lambda: value
        if 'w' in mode:
            fset = lambda value: self.symtab.install_symbol(node.name, value)
            
        return property(fget, fset)
    
    def accept_FieldAccessNode(self, node, mode='r'):
        assert isinstance(node, ast.FieldAccessNode)
        
        variable = self.accept(node.variable).fget()
        
        fget = fset = None
        if 'r' in mode:
            fget = lambda: getattr(variable, node.field)    
        if 'w' in mode:
            fset = lambda value: setattr(variable, node.field, value)
            
        return property(fget, fset)
    
    def accept_SubstitutionVariableNode(self, node):
        assert isinstance(node, ast.SubstitutionVariableNode)
        
        value = self.accept(node.expr).fget()
        value = self.runtime.format_string(value, node.formats)

        return property(lambda: value)
    
    def accept_SubstitutionNavigationNode(self, node):
        assert isinstance(node, ast.SubstitutionNavigationNode)
        
        variable = self.accept(node.variable).fget()
        chain = self.runtime.chain(variable)
        
        key_letter = node.navigation.key_letter
        rel_id = node.navigation.relation.rel_id
        phrase = node.navigation.relation.phrase
        
        inst_set = chain.nav(key_letter, rel_id, phrase)()
        value = self.runtime.select_any_in(inst_set, lambda selected: True)
        
        return property(lambda: value)
    
    def accept_ParseKeywordNode(self, node):
        assert isinstance(node, ast.ParseKeywordNode)
        
        keyword = self.accept(node.keyword).fget()
        value = self.accept(node.expr).fget()
        value = self.runtime.parse_keyword(value, keyword)
        
        return property(lambda: value)
    
    def accept_PrintNode(self, node):
        assert isinstance(node, ast.PrintNode)
        
        value = self.accept(node.value_list).fget()
        self.runtime.invoke_print(value)

    def accept_ExitNode(self, node):
        assert isinstance(node, ast.ExitNode)
        
        value = self.accept(node.return_code).fget()
        self.runtime.invoke_exit(value)
        
    def accept_IfNode(self, node):
        assert isinstance(node, ast.IfNode)
        
        try:
            self.symtab.enter_block()
            if self.accept(node.cond).fget():
                self.accept(node.iftrue)
            elif not self.accept(node.elif_list).fget():
                self.accept(node.iffalse)
            self.symtab.leave_block()
        except BreakException as e:
            self.symtab.leave_block()
            raise e
                        
    def accept_ElIfListNode(self, node):
        assert isinstance(node, ast.ElIfListNode)
        
        b = False
        for _elif in node.elifs:
            b = self.accept(_elif).fget()
            if b: break
            
        return property(lambda: b)
    
    def accept_ElIfNode(self, node):
        assert isinstance(node, ast.ElIfNode)
        
        b = self.accept(node.cond).fget()
        if b:
            self.accept(node.statement_list)

        return property(lambda: b)
    
    def accept_WhileNode(self, node):
        assert isinstance(node, ast.WhileNode)
        
        try:
            self.symtab.enter_block()
            while self.accept(node.cond).fget():
                self.accept(node.statement_list)
            self.symtab.leave_block()
        except BreakException:
            self.symtab.leave_block()
        
    def accept_ForNode(self, node):
        assert isinstance(node, ast.ForNode)
        
        try:
            self.symtab.enter_block()
            for value in iter(self.symtab.find_symbol(node.set_name)):
                self.symtab.install_symbol('selected', value)
                self.symtab.install_symbol(node.variable_name, value)
                self.accept(node.statement_list)
            self.symtab.leave_block()
        except BreakException:
            self.symtab.leave_block()

    def accept_BreakNode(self, node):
        assert isinstance(node, ast.BreakNode)
        
        raise BreakException()

    def accept_BinaryOpNode(self, node):
        assert isinstance(node, ast.BinaryOpNode)
        
        ops = {
            '|':   lambda lhs, rhs: (lhs | rhs),
            '&':   lambda lhs, rhs: (lhs & rhs),
            '+':   lambda lhs, rhs: (lhs + rhs),
            '-':   lambda lhs, rhs: (lhs - rhs),
            '*':   lambda lhs, rhs: (lhs * rhs),
            '/':   lambda lhs, rhs: (lhs / rhs),
            '<':   lambda lhs, rhs: (lhs < rhs),
            '<=':  lambda lhs, rhs: (lhs <= rhs),
            '>':   lambda lhs, rhs: (lhs > rhs),
            '>=':  lambda lhs, rhs: (lhs >= rhs),
            '!=':  lambda lhs, rhs: (lhs != rhs),
            '==':  lambda lhs, rhs: (lhs == rhs),
            'or':  lambda lhs, rhs: (lhs or  rhs),
            'and': lambda lhs, rhs: (lhs and rhs),
        }

        lhs = self.accept(node.left).fget()
        rhs = self.accept(node.right).fget()
        
        if node.sign in ['|', '&']:
            lhs = self.runtime.cast_to_set(lhs)
            
        if isinstance(lhs, xtuml.model.QuerySet):
            rhs = self.runtime.cast_to_set(rhs)
            
        value = ops[node.sign](lhs, rhs)
        
        return property(lambda: value)
    
    def accept_UnaryOpNode(self, node):
        assert isinstance(node, ast.UnaryOpNode)
        
        ops = {
            '-':           lambda value:-value,
            'not':         lambda value: not value,
            'cardinality': lambda value: self.runtime.cardinality(value),
            'empty':       lambda value: self.runtime.empty(value),
            'first':       lambda value: self.runtime.first(self.symtab.find_symbol('selected'), value),
            'last':        lambda value: self.runtime.last(self.symtab.find_symbol('selected'), value),
            'not_empty':   lambda value: self.runtime.not_empty(value),
            'not_first':   lambda value: self.runtime.not_first(self.symtab.find_symbol('selected'), value),
            'not_last':    lambda value: self.runtime.not_last(self.symtab.find_symbol('selected'), value),
        }

        value = self.accept(node.value).fget()
        value = ops[node.sign](value)
        
        return property(lambda: value)
    
    def accept_LiteralNode(self, node):
        assert isinstance(node, ast.LiteralNode)
        
        s = node.value
        
        return property(lambda: s)
    
    def accept_LiteralListNode(self, node):
        assert isinstance(node, ast.LiteralListNode)
        
        s = ''
        for literal in node.literals:
            s += self.accept(literal).fget()
         
        self.runtime.buffer_literal(s)
        
    def accept_EmitNode(self, node):
        assert isinstance(node, ast.EmitNode)
        
        filename = self.accept(node.emit_filename).fget()
        
        self.runtime.emit_buffer(filename)
    
    def accept_ClearNode(self, node):
        assert isinstance(node, ast.ClearNode)
        
        self.runtime.clear_buffer()
            
    def accept_IncludeNode(self, node):
        assert isinstance(node, ast.IncludeNode)
        
        filename = self.accept(node.inc_filename).fget()
        root = None
        
        # check cache
        if filename in self.runtime.include_cache:
            root = self.runtime.include_cache[filename]
        
        # check absolute path
        elif os.path.isabs(filename):
            root = parse.parse_file(filename)
            self.runtime.include_cache[filename] = root

        # search relative include paths
        else:
            paths_to_search = [os.path.dirname(self.runtime.info.arch_file_name)]
            paths_to_search.extend(self.includes)
            paths_to_search = filter(None, paths_to_search)
            
            for path in paths_to_search:
                abs_path = '%s/%s' % (path, filename)
                if os.path.exists(abs_path):
                    root = parse.parse_file(abs_path)
                    self.runtime.include_cache[filename] = root
                    break
            
        if root is None:
            raise Exception("unable to find '%s'" % filename)
        
        self.callstack.append(node)
        self.accept(root)
        self.callstack.pop()
        
    def accept_CreateNode(self, node):
        assert isinstance(node, ast.CreateNode)
        
        inst = self.runtime.new(node.key_letter)
        self.symtab.install_symbol(node.variable_name, inst)
        
        return property(lambda: inst)
        
    def accept_SelectAnyInstanceNode(self, node):
        assert isinstance(node, ast.SelectAnyInstanceNode)
        
        where = self.accept(node.where)
        value = self.runtime.select_any_from(node.key_letter, where)
        
        self.symtab.install_symbol(node.variable_name, value)
        
        return property(lambda: value)
        
    def accept_SelectManyInstanceNode(self, node):
        assert isinstance(node, ast.SelectManyInstanceNode)

        where = self.accept(node.where)
        value = self.runtime.select_many_from(node.key_letter, where)
        
        self.symtab.install_symbol(node.variable_name, value)
        
        return property(lambda: value)

    def accept_SelectOneNode(self, node):
        assert isinstance(node, ast.SelectOneNode)
        
        inst_set = iter(self.accept(node.instance_chain).fget())
        where = self.accept(node.where)
        value = self.runtime.select_one_in(inst_set, where)
        
        self.symtab.install_symbol(node.variable_name, value)
        
        return property(lambda: value)

    def accept_SelectAnyNode(self, node):
        assert isinstance(node, ast.SelectAnyNode)
        
        inst_set = iter(self.accept(node.instance_chain).fget())
        where = self.accept(node.where)
        value = self.runtime.select_any_in(inst_set, where)
        
        self.symtab.install_symbol(node.variable_name, value)
        
        return property(lambda: value)
        
    def accept_SelectManyNode(self, node):
        assert isinstance(node, ast.SelectManyNode)
        
        inst_set = iter(self.accept(node.instance_chain).fget())
        where = self.accept(node.where)
        value = self.runtime.select_many_in(inst_set, where)
        
        self.symtab.install_symbol(node.variable_name, value)
        
        return property(lambda: value)

    def accept_WhereNode(self, node):
        assert isinstance(node, ast.WhereNode)
        
        def where(expr, selected):
            if node.expr is None: return True
            
            self.symtab.enter_block()
            self.symtab.install_symbol('selected', selected)
            value = self.accept(expr).fget()
            self.symtab.leave_block()
            
            return value
            
        return lambda selected: where(node.expr, selected)
        
    def accept_InstanceChainNode(self, node):
        assert isinstance(node, ast.InstanceChainNode)
        
        inst = self.accept(node.variable).fget()
        chain = self.runtime.chain(inst)
        
        for nav in node.navigations:
            rel_id = nav.relation.rel_id
            phrase = nav.relation.phrase
            chain = chain.nav(nav.key_letter, rel_id, phrase)
        
        result = chain()
        
        return property(lambda: result)
    
    
def evaluate(rt, ast, includes):
    w = EvalWalker(rt, includes)
    w.accept(ast)
    



# encoding: utf-8
# Copyright (C) 2015 John Törnblom
'''
Symbol table for the rule-specification language (RSL).  
'''


class SymtabException(Exception):
    pass


class Block(dict):
    pass


class Scope(list):
    
    def __init__(self):
        self.append(Block())

    @property
    def symbols(self):
        return_symbols = dict()
        for d in self:
            return_symbols.update(d)
            
        return return_symbols.items()
    

class SymbolTable(object):

    def __init__(self):
        self._global_scope = dict()
        self._functions = dict()
        self._scopes = list()
        
    @property
    def scope_head(self):
        if not len(self._scopes): 
            raise SymtabException('Out of scope')
        
        return self._scopes[-1]
    
    def enter_scope(self):
        scope = Scope()
        self._scopes.append(scope)

    def leave_scope(self):
        if not len(self._scopes): 
            raise SymtabException('Out of scope')
        
        d = dict()
        for block in self._scopes.pop():
            d.update(block)
        return d
    
    def enter_block(self):        
        block = Block()
        self.scope_head.append(block)
    
    def leave_block(self):
        if not len(self.scope_head): 
            raise SymtabException('Out of block')
        
        block = self.scope_head.pop()
        del block
    
    def install_global(self, name, handle):
        self._global_scope[name] = handle
        
    def install_symbol(self, name, handle):
        for block in self.scope_head:
            for key in block.keys():
                if key.lower() == name.lower():
                    block[key] = handle
                    return

        block = self.scope_head[-1]
        block[name] = handle

    def find_symbol(self, name):
        for block in self.scope_head:
            for key in block.keys():
                if key.lower() == name.lower():
                    return block[key]
        
        for key in self._global_scope.keys():
            if key.lower() == name.lower():
                return self._global_scope[key]
        
        raise SymtabException("Unknown symbol '%s'" % name)





# encoding: utf-8
# Copyright (C) 2015 John Törnblom
'''
High-level runtime behavior for the RSL language, e.g. builtin functions and 
helper functions like 'emit to file'.
'''


import sys
import os
import stat
import subprocess
import datetime
import logging
import re
import difflib

import rsl.version
import xtuml.model


try:
    from future_builtins import filter
except ImportError:
    pass


logger = logging.getLogger(__name__)


class RuntimeException(Exception):
    pass


class Info(object):
    '''
    Helper class for providing access to the built-in
    substitution variables "${info.date}" et.al.
    '''
    def __init__(self, metamodel):
        self.metamodel = metamodel
        self.arch_file_name = ''
        self.arch_file_line = 0
        
    @property
    def date(self):
        now = datetime.datetime.now()
        now = datetime.datetime.ctime(now)
        return now
    
    @property
    def unique_num(self):
        return next(self.metamodel.id_generator)
    
    @property
    def user_id(self):
        return os.getlogin()
    
    @property
    def interpreter_version(self):
        return rsl.version.complete_string
    
    @property
    def interpreter_platform(self):
        return os.name


class Fragment(xtuml.model.BaseObject):
    __r__ = dict()
    __q__ = dict()
    __c__ = dict()
    
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)
        xtuml.model.BaseObject.__init__(self)


class Runtime(object):

    def __init__(self, metamodel, emit=None, force=False, diff=None):
        self.metamodel = metamodel
        self.emit = emit
        self.force_emit = force
        self.diff = diff
        self.functions = dict()
        self.buffer = ''
        self.include_cache = dict()
        self.info = Info(metamodel)
            
        self.define_function('GET_ENV_VAR', self.get_env_var)
        self.define_function('PUT_ENV_VAR', self.put_env_var)
        self.define_function('SHELL_COMMAND', self.shell_command)
        self.define_function('FILE_READ', self.file_read)
        self.define_function('FILE_WRITE', self.file_write)
        self.define_function('STRING_TO_INTEGER', self.string_to_integer)
        self.define_function('STRING_TO_REAL', self.string_to_real)
        self.define_function('INTEGER_TO_STRING', self.integer_to_string)
        self.define_function('REAL_TO_STRING', self.real_to_string)
        self.define_function('BOOLEAN_TO_STRING', self.boolean_to_string)
        
    @staticmethod        
    def get_env_var(name):
        if name in os.environ:
            result = os.environ[name]
            success = True
        else:
            result = ''
            success = False
        
        return {'attr_success': success,
                'attr_result': result}
    
    @staticmethod
    def put_env_var(value, name):
        os.environ[name] = value
        return {'attr_success': name in os.environ}
    
    @staticmethod
    def shell_command(cmd):
        return {'attr_result': subprocess.call(cmd, shell=True)}
    
    @staticmethod
    def file_read(filename):
        try:
            with open(filename, 'r') as f:
                result = f.read()
                success = True
        except:
            success = False
            result = ''
            
        return {'attr_success': success,
                'attr_result': result}
    
    @staticmethod
    def file_write(contents, filename):
        try:
            with open(filename, 'w') as f:
                f.write('%s\n' % contents)
                success = True
        except:
            success = False
        
        return {'attr_success': success}
        
    @staticmethod
    def string_to_integer(value):
        return {'attr_result': int(value)}
    
    @staticmethod
    def string_to_real(value):
        return {'attr_result': float(value)}
        
    @staticmethod
    def integer_to_string(value):
        return {'attr_result': str(value)}
        
    @staticmethod
    def real_to_string(value):
        return {'attr_result': str(value)}
        
    @staticmethod
    def boolean_to_string(value):
        return {'attr_result': str(value).upper()}    
        
    @staticmethod
    def format_string(expr, fmt):
        whitespace_regexp = re.compile(r'\s+')
        nonword_regexp = re.compile(r'[^\w]')
            
        def o(value):
            value = value.replace('_', ' ')
            value = value.title()
            value = re.sub(nonword_regexp, '', value)
            value = re.sub(whitespace_regexp, '', value)
            if value:
                value = value[0].lower() + value[1:]
            return value
            
        ops = {
                'u' : lambda value: value.upper(),
                'c' : lambda value: value.title(),
                'l' : lambda value: value.lower(),
                '_' : lambda value: re.sub(whitespace_regexp, '_', value),
                'r' : lambda value: re.sub(whitespace_regexp, '', value),
                't' : lambda value: value,
                'o' : o
        }
        
        s = '%s' % expr
        for ch in fmt:
            s = ops[ch.lower()](s)
    
        return s
    
    @staticmethod
    def parse_keyword(expr, keyword):
        regexp = re.compile(keyword + ":([^\n]*)")
        result = regexp.search(expr)
        
        if result:
            return result.groups()[0]
        else:
            return ''
    
    def define_function(self, name, fn):
        self.functions[name] = fn
        
    def invoke_function(self, name, args):
        if name not in self.functions:
            raise RuntimeException("Function '%s' is undefined" % name)
        
        buf = self.clear_buffer()
        
        fn = self.functions[name]
        d = fn(*args)
        
        return_values = dict({'body': self.clear_buffer()})
        for key, value in d.items():
            if key.startswith("attr_"):
                key = key.split("attr_")[1]
                return_values[key] = value
    
        self.buffer = buf
        
        return Fragment(**return_values)
    
    @staticmethod
    def invoke_print(value):
        sys.stderr.write("%s\n" % value)
    
    @staticmethod
    def invoke_exit(exit_code):
        sys.exit(exit_code)
        
    @staticmethod
    def cast_to_set(value):
        if not isinstance(value, xtuml.model.QuerySet):
            return xtuml.model.QuerySet([value])
        else:
            return value
        
        
    def buffer_literal(self, literal):
        if   literal.endswith('\\' * 3):
            self.buffer += literal[:-2]
        
        elif literal.endswith('\\' * 2):
            self.buffer += literal[:-1] + '\n'
            
        elif literal.endswith('\\'):
            self.buffer += literal[:-1]
            
        elif literal.endswith('\n'):
            self.buffer += literal
            
        else:
            self.buffer += literal + '\n'
    
    def append_diff(self, filename, org, buf):
        org = org.splitlines(1)
        buf = buf.splitlines(1)
        
        fromfile = filename
        tofile = filename
        
        if os.path.exists(filename):
            fromdate = os.path.getctime(filename)
            fromdate = datetime.datetime.fromtimestamp(fromdate)
            todate = str(datetime.datetime.now())
        else:
            fromdate = ''
            todate = ''
        
        diff = difflib.unified_diff(org, buf, fromfile, tofile, fromdate, todate)

        with open(self.diff, 'a') as f:
            f.write(''.join(diff))

    def emit_buffer(self, filename):
        org = ''
        buf = self.clear_buffer()
        if buf and not buf.endswith('\n'):
            buf += '\n'
            
        filename = os.path.normpath(filename)
        if os.path.exists(filename):
            with open(filename, 'rU') as f:
                org = f.read()
        
        if self.emit == 'never':
            do_write = False 
            
        elif self.emit == 'change' and org == buf:
            do_write = False
        
        else:
            do_write = True
                
        if self.diff:
            self.append_diff(filename, org, buf)

        if do_write and self.force_emit and os.path.exists(filename):
            st = os.stat(filename)
            os.chmod(filename, st.st_mode | stat.S_IWRITE)

        if do_write:
            dirname = os.path.dirname(filename)
            if dirname and not os.path.exists(dirname):
                os.makedirs(dirname)

            if os.path.exists(filename):
                logger.info("File '%s' REPLACED" % filename)
            else:
                logger.info("File '%s' CREATED" % filename)

            with open(filename, 'w+') as f:
                f.write(buf)

    def clear_buffer(self):
        b = self.buffer
        self.buffer = ''
        return b
    
    def new(self, key_letter):
        return self.metamodel.new(key_letter)
    
    def chain(self, inst):
        return self.metamodel.navigate_many(inst)
    
    def select_any_from(self, key_letter, where_cond):
        return self.metamodel.select_any(key_letter, where_cond)
         
    def select_many_from(self, key_letter, where_cond):
        return self.metamodel.select_many(key_letter, where_cond)

    @staticmethod
    def select_many_in(inst_set, where_cond):
        s = filter(where_cond, inst_set)
        return xtuml.model.QuerySet(s)

    @staticmethod
    def select_any_in(inst_set, where_cond):
        for inst in iter(inst_set):
            if where_cond(inst):
                return inst

    @staticmethod
    def select_one_in(inst_set, where_cond):
        cardinality = xtuml.model.MetaModel.cardinality(inst_set)
        if cardinality > 1:
            raise RuntimeException('select one from a set with cardinality %d' % cardinality)
        
        return Runtime.select_any_in(inst_set, where_cond)
                
    @staticmethod
    def cardinality(inst_set):
        return xtuml.model.MetaModel.cardinality(inst_set)
    
    @staticmethod
    def empty(inst_set):
        return xtuml.model.MetaModel.empty(inst_set)
    
    @staticmethod
    def not_empty(inst_set):
        return xtuml.model.MetaModel.not_empty(inst_set)
    
    @staticmethod
    def first(inst, inst_set):
        return xtuml.model.MetaModel.first(inst, inst_set)
    
    @staticmethod
    def not_first(inst, inst_set):
        return xtuml.model.MetaModel.not_first(inst, inst_set)
    
    @staticmethod
    def last(inst, inst_set):
        return xtuml.model.MetaModel.last(inst, inst_set)
    
    @staticmethod
    def not_last(inst, inst_set):
        return xtuml.model.MetaModel.not_last(inst, inst_set)
    
    @staticmethod
    def is_set(inst):
        return xtuml.model.MetaModel.is_set(inst)

    @staticmethod
    def is_instance(inst):
        return xtuml.model.MetaModel.is_instance(inst)

    def assert_type(self, exptected_type, value):
        value_type = self.type_name(type(value))
        if exptected_type != value_type:
            raise RuntimeException('expected type %s, not %s' % (exptected_type, value_type))
        
    def type_name(self, ty):
        if issubclass(ty, Fragment):
            return 'frag_ref'
        else: 
            return self.metamodel.type_name(ty)
    
    def named_type(self, name):
        if name == 'frag_ref':
            return Fragment
        else: 
            return self.metamodel.named_type(name)
        



# encoding: utf-8
# Copyright (C) 2015 John Törnblom
'''
Keep version info in one single place
'''


name = 'pyrsl'
date = '2015-05-05'
release = '0.3.1'

complete_string = '%s v%s (%s)' % (name, release, date)
